/*
 *   Copyright 2009 Joubin Houshyar
 * 
 *   Licensed under the Apache License, Version 2.0 (the "License");
 *   you may not use this file except in compliance with the License.
 *   You may obtain a copy of the License at
 *    
 *   http://www.apache.org/licenses/LICENSE-2.0
 *    
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 */

package org.jredis.ri.alphazero.wip;

import java.io.IOException;
import java.io.InputStream;

/**
 * A reusable, auto-resizing, (byte array wrapping) data buffer.  
 * <b>This class is not thread safe!</b>
 * 
 * @author Joubin Houshyar (alphazero@sensesay.net)
 *
 */
/**
 * [TODO: document me!]
 *
 * @author  Joubin Houshyar (alphazero@sensesay.net)
 * @version alpha.0, Apr 29, 2009
 * @since   alpha.0
 * 
 */
@Deprecated
public class DataBuffer {

	// ------------------------------------------------------------------------
	// Fields and attributes
	// ------------------------------------------------------------------------
	/** */
	public static final int DEFAULT_GROW_LEN = 1024;

	/** */
	private static final String EMPTY_STRING = "";

	/** */
	private byte[]	buffer = null;
	
	/** Write Offset */
	private int		offset = 0; // points to the next available buffer index 
	
	/** */
	private final int optimalSize;
	
	/** */
	private final int grow_len;

	// ------------------------------------------------------------------------
	// Constructors
	// ------------------------------------------------------------------------
	/**
	 * Default grow length of 1024
	 * @param optimalSize used both for initial allocation and subsequent resets.
	 */
	public DataBuffer (int optimalSize) {
		this(optimalSize, DEFAULT_GROW_LEN);
	}

	/**
	 * 
	 * @param optimalSize
	 * @param growLen
	 */
	public DataBuffer (int optimalSize, int growLen) {
		this.optimalSize = optimalSize;
		this.grow_len = growLen;
		buffer = new byte [optimalSize];
		
	}

	// ------------------------------------------------------------------------
	// INTERFACE -- (non-destructive) informational methods
	// ------------------------------------------------------------------------
	/** 
	 * returns the length of data written. (This is <b>not</b> the size of the underlying
	 * buffer.)
	 * <p><code>0 <= length() <= size()</code> always holds true.
	 * <p>i.e. <code>dataBuffer.getData (false).length == dataBuffer.length()</code>
	 */
	public int offset () {return offset;}

	/** 
	 * returns the size of the data buffer.  
	 * This value is equal or greater than the length of the data and can only change by 
	 * (internal) calls to {@link DataBuffer#grow(int)} 
	 */
	public int size () { return buffer.length;}

	/**
	 * @return the number of available bytes that can be written to this buffer without
	 * resizing the underlying byte[] array.
	 */
	public int available () { return  buffer.length - offset; }

	/**
	 * @return the optimal size of this buffer.  
	 */
	public int getOptimalSize() { return optimalSize; }
	
	// ------------------------------------------------------------------------
	// INTERFACE -- public - meta-data modifying methods
	// ------------------------------------------------------------------------
	/**
	 * Resets the buffer offset -- will not erase the data already written (if any) so that
	 * subsequent call to {@link DataBuffer#getBytes()} will contain the data that was previously written.
	 * 
	 * @return
	 */
	public DataBuffer reset() {
		offset = 0;
		return this;
	}
	/**
	 * Resets the buffer offset as per {@link DataBuffer#reset()}.  Additionally, it
	 * will check to see if the current buffer is larger than the optimal size, and if so,
	 * it will discard the buffer to be garbage collected and re-allocates a new optimally sized buffer.
	 * @return
	 */
	public DataBuffer reset (boolean optimize) {
		reset();
		if(size() > optimalSize) buffer = new byte[optimalSize];

		return this;
	}
	
	/** effectively rewinds the offset (i.e. current index) by cnt,
	 * if offset is NOT less than cnt.   
	 * @param cnt
	 * @return true if did it, or false if it could not.
	 */
	public boolean trim(int cnt) {
		if(offset >= cnt) {
			offset -= cnt;
			return true;
		}
		return false;
	}

	// ------------------------------------------------------------------------
	// INTERFACE -- public - data access methods
	// ------------------------------------------------------------------------
	/**
	 * Returns the byte data at the given index.
	 * @see DataBuffer#offset
	 * 
	 * @param i index value, from zero up to offset.
	 * @throws ArrayIndexOutOfBoundsException if index is negative, or equal or
	 * greater than the offset.
	 * 
	 * @return the byte value at the index.
	 */
	public byte get(int i) {
		if(i<0) 
			throw new ArrayIndexOutOfBoundsException("bad index: " + i);
		if(i>=this.offset) 
			throw new ArrayIndexOutOfBoundsException ("bad index: " + i + " is past the current offset of " + offset);
		
		return this.buffer[i];
	}
	/**
	 * Gets the data, and optionally resets the buffer.
	 * This method incurs the cost of (a) allocating a buffer to hold the data. 
	 * @return a byte buffer with size corresponding to the value of {@link DataBuffer#offset()};
	 */
	public byte[] getData (boolean reset) {
		byte[] data = new byte[offset];
		System.arraycopy(buffer, 0, data, 0, offset);
		if(reset) reset();
		return data;
	}

	/**
	 * Gets the data but <b>does *not* reset the offset</b>.  Explicitly, think of this method
	 * as get-data-as-of-now.  Subsequent calls to the buffer with append, for example, 
	 * will continue appending at the point you were before you made this call.
	 * 
	 * This method incurs the cost of (a) allocating a buffer to hold the data. 
	 * @return a byte buffer with size corresponding to the value of {@link DataBuffer#offset()};
	 */
	public byte[] getData () {
		byte[] data = new byte[offset];
		System.arraycopy(buffer, 0, data, 0, offset);
		return data;
	}

	/**
	 * Faster than calling getData and then making a string out of it ...
	 * Remember, doesn't reset this buffer for you!
	 * <p>This routine is guaranteed to never return a null reference, but
	 * it will return a {@link String} object of length 0. (i.e. "") if the buffer
	 * offset is <code>0</code>.
	 * @return 
	 */
	public String toString (boolean reset) {
		if(offset == 0) return EMPTY_STRING;
		return new String (buffer, 0, offset);
	}

	/** 
	 * this simply returns the data buffer, which certainly may include garbage data
	 * after the current offset, but it is faster than getData() method as it will <b>not</b>
	 * incur the cost of {@link System#arraycopy(Object, int, Object, int, int)}.  
	 * 
	 * <p>Use it in  conjunction with getLength(), if you must, but note that if you write
	 * directly to the returned array, this class will not be tracking the shift in offset.
	 * @return
	 */
	public byte[] getBytes () {
		return buffer;
	}

	// ------------------------------------------------------------------------
	// INTERFACE -- public - data write methods
	// ------------------------------------------------------------------------
	/**
	 * Appends the byte and increments the internal offset.
	 * @param data
	 * @return
	 */
	public DataBuffer append(byte data) {
		//			int available = buffer.length - offset;
		if(buffer.length == offset) 							/* <--------------- available compute */
			grow(1);
		buffer[offset++] = data;
		return this;
	}
	/**
	 * Appends the data.  will grow buffer as necessary.  Internal offset will be incremented by <code>data.length</code>
	 * @param data
	 * @return
	 */
	public DataBuffer append (byte[] data){
		int len = data.length;
		int available = buffer.length - offset; 				/* <--------------- available compute */
		if(len > available ) 
			grow (len - available);
		System.arraycopy(data, 0, buffer, offset, len);
		offset += len;
		return this;
	}
	/**
	 * Appends the data.  will grow buffer as necessary.  Internal offset will be incremented by <code>data.length</code>
	 * @param data
	 * @return
	 */
	public DataBuffer append (byte[] data, int off, int len){
		int available = buffer.length - offset;					/* <--------------- available compute */
		if(len > available ) 
			grow (len - available);
		System.arraycopy(data, off, buffer, offset, len);
		offset += len;
		return this;
	}


	/**
	 * (Attempts to) read and append (at most) length bytes of data from the inputstream provided.
	 * Will initialy grow this buffer to meet the expected maximum requirements.
	 * 
	 * @param in
	 * @param length
	 * @return the number of bytes actually read and appended.  Value can be -1 (if nothing was read),
	 * or 0 or more (up to max).  
	 * @throws IOException 
	 * @throws IOException propagated from underlying stream.
	 */
	public int append(InputStream in, int max) throws IOException {
		int available;					
		available = buffer.length - offset;			/* <--------------- available compute */
		if(available < max) grow (max);
		int c=-1;
		if((c = in.read(buffer, offset, max)) < 0) return c;
		offset += c;
		return c;
	}

	
	// ------------------------------------------------------------------------
	// Inner ops.
	// ------------------------------------------------------------------------
	/**
	 * This is a somewhat costly op as it will definitively incur the cost of reallocating a larger byte buffer
	 * and performing a {@link System#arraycopy(Object, int, Object, int, int)} to transfer existing
	 * data from the old buffer.
	 * 
	 * <p>Buffer can certainly grow (much) larger than the optimal size specified, and if you do not wish
	 * to hold on to the larger buffer after you are finished with your data transferring requirements, it
	 * is recommended to call {@link DataBuffer#reset(boolean)} with <code>true</code>.
	 * 
	 * @param min the minimum amount that we need to grow the buffer, in bytes.
	 */
	private final void grow (int min) {	
		int growby = grow_len < min ? min : grow_len;
		byte[] newbuffer = new byte [buffer.length + growby];
		System.arraycopy(buffer, 0, newbuffer, 0, offset);
		buffer = newbuffer;
		//			size = buffer.length;
	}

	// ------------------------------------------------------------------------
	// deprecated stuff -- may be resurrected, so ...
	// ------------------------------------------------------------------------

//	/**
//	 * Reads a sequence of bytes, up to and including a sequence of bytes.
//	 * This is a rather limited method in that it expects the terminal sequence
//	 * to be both uniquely distinct within the stream, <b>AND</b>, that the
//	 * initial sequence byte <b>ONLY</b> occur in the terminal sequence.
//	 * 
//	 * <p>typical use would be to read a CRLF or (byte) 0 sequences of bytes.
//	 * 
//	 * @param in
//	 * @param terminalSequence
//	 * @return total bytes read.  (buffer.getOffset() before and after this method
//	 * has successfully run will be equal to the returned results.
//	 * 
//	 * @throws IllegalStateException if terminal sequence is never reached.
//	 * 
//	 * @throws IllegalArgumentException if terminal sequence provided is null or zero member array
//	 * 
//	 * @throws IOException (propagated) from the provided {@link InputStream}
//	 * if something goes wrong.  The data up to a point would have
//	 * been written to the buffer, but the offset will remain unaffected.  (You can
//	 * potentially see what was written by accessing {@link DataBuffer#getBytes()}
//	 */
//	public  
//	int append (InputStream in, byte[] terminalSequence) 
//		throws IOException, IllegalArgumentException, IllegalStateException
//	{
//		if(null == terminalSequence) throw new IllegalArgumentException ("null terminalSequence!");
//
//		int termlen = terminalSequence.length;
//		if(termlen == 0) throw new IllegalArgumentException ("zero length terminalSequence!");
//
//		int available;					
//		available = buffer.length - offset;			/* <--------------- available compute */
//		if(available < 128) grow (128);
//		
//		int mark=offset;  // we note our (initial) current offset
//		while (true){
//			int c=-1;
//			if((c = in.read (buffer, offset, 1)) < 1) {
//				int readtot = offset - mark;
//				offset = mark;  // roll back - reset the buffer offset
//				if(c==-1) throw new IOException("Unexpected EOF -- read return -1.  readtot: " + readtot);
//				throw new IllegalStateException ("expecting more data here! read tot: " + readtot + " in.read(): " + c);
//			}
//			
//			offset++;
//			
//			// we basically make sure that we have enough room to read the terminal
//			// sequence assuming that the next time round we'll find it.
//			// this way we'll never exceed buffer bounds and we can keep the code simple.
//			available = buffer.length - offset;
//			if(available < termlen) grow (termlen);
//			
//			if (buffer[offset-1]==terminalSequence[0]) {
//				--termlen;
//				if((in.read (buffer, offset, termlen)) < termlen) { 
//					throw new RuntimeException ("expecting more data here ... attempting to read " +
//							                     "the remaining elements of terminal sequence");
//				}
//				
//				// we either should have the terminal sequence from offset to offset+terminalSequence.length
//				for(int j=0; j<=termlen; j++) {
//					if(buffer[offset+j-1]!=terminalSequence[j]) {
//						offset = mark;
//						throw new IllegalStateException ("Expecting the terminal sequence read -- " +
//														"read data and sequence differ beginning at idx " + j + 
//														" -- Buffer offset reset to initial position.");
//					}
//				}
//				// looks ok
//				offset += termlen;
//				break;
//			}
//		}
//		return offset-mark;
//	}
	
}
